import torch
from sparseprop.modules.linear import SparseLinear
from sparseprop.modules.conv2d import SparseConv2d
from sparseprop.modules import sparsify_if_faster

@torch.no_grad()
def sparsity(module):
    if hasattr(module, 'weight'):
        return 1. - torch.mean((module.weight != 0).float()).item()
    else:
        return 0.

def swap_module(network, module_name, new_module):
    name_parts = module_name.split('.')
    parent = network
    for part in name_parts[:-1]:
        if part.isdigit():
            parent = parent[int(part)]
        else:
            parent = getattr(parent, part)
    
    last_part = name_parts[-1]
    if last_part.isdigit():
        parent[int(last_part)] = new_module
    else:
        setattr(parent, last_part, new_module)

class ShapeHook:
    def __init__(self):
        self.inshape = None
        self.outshape = None
        
    def __call__(self, module, inp, out):
        if isinstance(inp, tuple):
            inp = inp[0]
        if isinstance(out, tuple):
            out = out[0]
        
        self.inshape = inp.shape[1:]
        self.outshape = out.shape[1:]

def generate_intermediate_shapes(network, input_shape):
    hooks = {}
    handles = []
    for name, module in network.named_modules():
        if any([isinstance(module, c) for c in [torch.nn.Linear, torch.nn.Conv2d]]):
            hook = ShapeHook()
            handles.append(module.register_forward_hook(hook))
            hooks[name] = hook

    B = 1
    training_mode = network.training
    network.eval()
    with torch.no_grad():
        network(torch.randn(B, *input_shape))
    network.train(training_mode)
    
    inshapes = {name: hook.inshape for name, hook in hooks.items()}
    outshapes = {name: hook.outshape for name, hook in hooks.items()}

    for handle in handles:
        handle.remove()
    
    return inshapes, outshapes

def swap_modules_with_sparse(network, input_shape, inplace=False, skip_modules=None, verbose=False):
    # e.g., shapes_tag='resnet18', input_shape=(B, IC, M, N), skip_modules='input,conv1,conv2'

    if not inplace:
        network = deepcopy(network)
    if skip_modules is not None:
        skip_modules = skip_modules.split(',')

    B = input_shape[0]
    input_shape = input_shape[1:]
    inshapes, _ = generate_intermediate_shapes(network, input_shape)

    for name, module in network.named_modules():
        is_conv = isinstance(module, torch.nn.Conv2d)
        is_linear = isinstance(module, torch.nn.Linear)
        if not is_conv and not is_linear:
            continue
        
        found = False
        if skip_modules is not None:
            for sm in skip_modules:
                if name == sm.strip():
                    found = True
                    break
        if found:
            print(f'Skipped {name}.')
            continue
        
        sp = sparsity(module)
        new_module = None
        if sp > .8:
            new_module = sparsify_if_faster(
                module,
                (B, *inshapes[name]),
                verbose=verbose
            )

        if new_module is not None and new_module != module:
            swap_module(network, name, new_module)
            print(f'module {name} replaced with {str(new_module)}')
        else:
            print(f'keeping the module {name} dense...')

    return network

def error(pred, target):
    e = torch.mean((pred - target) ** 2) / torch.norm(target) ** 2
    return e.item()


import torch
from copy import deepcopy
import time
from scipy.sparse import csr_matrix
from sparseprop import backend as sppb

def to_csr_2d(data):
    if isinstance(data, torch.nn.Parameter):
        data = data.data
    spa = csr_matrix(data, shape=data.shape)
    val = torch.Tensor(spa.data)
    idx_N = torch.Tensor(spa.indptr).int()
    idx_M = torch.Tensor(spa.indices).int()
    return val, (idx_N, idx_M)

def to_sparse_format_conv2d(dense_weight):
    OC, IC, K, _ = dense_weight.shape
    nnz = torch.sum(dense_weight != 0).item()
    W_val = torch.zeros(nnz).float()
    W_OC = torch.zeros(OC + 1).int()  
    W_IC = torch.zeros((IC + 1) * OC).type(torch.short)
    W_X = torch.zeros(nnz).type(torch.uint8)
    W_Y = torch.zeros(nnz).type(torch.uint8)
    sppb.sparsify_conv2d(OC, IC, K, dense_weight.data, W_OC, W_IC, W_X, W_Y, W_val)
    return W_val, (W_OC, W_IC, W_X, W_Y)


import torch
from scipy.sparse import csr_matrix

from sparseprop.modules.functions import SparseLinearFunction
from sparseprop.modules.utils import to_csr_2d

class SparseLinear(torch.nn.Module):
    def __init__(self, dense_weight, bias=None):
        super(SparseLinear, self).__init__()
        self.N, self.M = dense_weight.shape

        W_val, W_idx = to_csr_2d(dense_weight)
        self.W_val = torch.nn.Parameter(W_val)
        self.W_idx = W_idx

        assert bias is None or isinstance(bias, torch.nn.Parameter), f"bias is not a parameter but it's {type(bias)}"
        self.bias = bias
    
    @staticmethod
    def from_dense(module):
        return SparseLinear(
            dense_weight=module.weight.data,
            bias=None if module.bias is None else torch.nn.Parameter(module.bias.data.clone())
        )

    @property
    def weight(self):
        return self.W_val
    
    def forward(self, input):
        return SparseLinearFunction.apply(input, self.W_val, self.W_idx, self.bias, self.N)

    @torch.no_grad()
    def apply_further_mask(self, new_mask):
        """
            This function is used when we need to further sparsify a sparse module, e.g., gradual pruning.
        """

        indptr, indices = self.W_idx
        dense_weight = torch.Tensor(csr_matrix((
            self.W_val.data, 
            indices, 
            indptr
        ), shape=(self.N, self.M)).toarray()).float()

        dense_mask = torch.Tensor(csr_matrix((
            new_mask, 
            indices, 
            indptr
        ), shape=(self.N, self.M)).toarray()).float()
        
        W_val, W_idx = to_csr_2d(dense_weight * dense_mask)
        self.W_val = torch.nn.Parameter(W_val)
        self.W_idx = W_idx

    def __str__(self):
        nnz = len(self.W_val)
        numel = self.N * self.N
        return f"SparseLinear([{self.N}, {self.M}], sp={1. - nnz/numel:.2f}, nnz={nnz})"


import torch
import time
from copy import deepcopy

from sparseprop.modules.conv2d import SparseConv2d
from sparseprop.modules.linear import SparseLinear

@torch.enable_grad()
def run_and_choose(modules, input_shape, verbose=False):
    if len(modules) == 1:
        if verbose:
            print('only one option...')
        return modules[0]

    X_orig = torch.randn(*input_shape)
    Y_orig = None

    min_time = 1e10
    best_module = None
    for module in modules:
        module_copy = deepcopy(module)
        X = X_orig.clone()
        X.requires_grad_()
        X.retain_grad()

        temp = time.time()
        O = module_copy(X)
        fwd_time = time.time() - temp

        if Y_orig is None:
            Y_orig = torch.randn_like(O)
        Y = Y_orig.clone()

        L = torch.mean((O - Y) ** 2)
        temp = time.time()
        L.backward()
        bwd_time = time.time() - temp

        if verbose:
            print(f'module {module} took {fwd_time} fwd and {bwd_time} bwd')

        full_time = fwd_time + bwd_time
        if full_time < min_time:
            min_time = full_time
            best_module = module
    
    if verbose:
        print(f'going with {best_module} with full time of {min_time}')
    return best_module

def _sparsify_if_faster_linear(module, input_shape, include_dense, verbose):
    sp = SparseLinear(
        dense_weight=module.weight.data,
        bias=None if module.bias is None else torch.nn.Parameter(module.bias.data.clone())
    )
    
    if not include_dense:
        return sp

    assert input_shape is not None
    return run_and_choose([module, sp], input_shape, verbose=verbose)

def _sparsify_if_faster_conv2d(conv, input_shape, include_dense, verbose):
    def bias_to_param():
        if conv.bias is None:
            return None
        return torch.nn.Parameter(conv.bias.data.clone())

    dense_weight = conv.weight.data
    stride = conv.stride[0]
    padding = conv.padding[0]

    sp1 = SparseConv2d(
        dense_weight,
        bias=bias_to_param(),
        padding=padding,
        stride=stride,
        vectorizing_over_on=False
    )
    
    sp2 = SparseConv2d(
        dense_weight,
        bias=bias_to_param(),
        padding=padding,
        stride=stride,
        vectorizing_over_on=True
    )

    modules = []
    if include_dense:
        modules.append(conv)
    modules += [sp1, sp2]
    
    return run_and_choose(modules, input_shape, verbose=verbose)

def sparsify_if_faster(module, input_shape, include_dense=True, verbose=False):
    if isinstance(module, torch.nn.Linear):
        return _sparsify_if_faster_linear(module, input_shape, include_dense, verbose)
    else:
        assert isinstance(module, torch.nn.Conv2d)
        return _sparsify_if_faster_conv2d(module, input_shape, include_dense, verbose)

def sparsify_conv2d_auto(conv, input_shape, verbose=False):
    return _sparsify_if_faster_conv2d(conv, input_shape, include_dense=False, verbose=verbose)

class TimingHook:
    def __init__(self, tag=None, verbose=False):
        self.clear()
        self._tag = tag
        self._verbose = verbose
        
    def __call__(self, module, inp, out):
        self.time = time.time()
        self.count += 1
        
        if isinstance(inp, tuple):
            inp = inp[0]
        if isinstance(out, tuple):
            out = out[0]

        if self._verbose:
            print(f"[Hook {self._tag}] inp: {inp.shape if inp is not None else inp}, out: {out.shape if out is not None else None}")
        
    def clear(self):
        self.time = None
        self.count = 0


def attach_identity_and_time(module, X, Y, time_forward=False, time_backward=True):
    if not time_backward:
        t = time.time()
        O = module(X)
        return time.time() - t
    else:
        identity = torch.nn.Identity()

        X.requires_grad_()
        X.retain_grad()

        module_hook = TimingHook()
        identity_hook = TimingHook()
        handles = [
            module.register_full_backward_hook(module_hook),
            identity.register_full_backward_hook(identity_hook)
        ]

        t = time.time()
        O_before_identity = module(X)
        forward_time = time.time() - t

        O = identity(O_before_identity)
        L = torch.mean((O - Y) ** 2)

        L.backward()

        backward_time = module_hook.time - identity_hook.time

        for handle in handles:
            handle.remove()
                
        if time_forward:
            return backward_time, forward_time
        else:
            return backward_time


import torch
from sparseprop.modules.utils import to_sparse_format_conv2d
from sparseprop.modules.functions import SparseConvFunction
from sparseprop import backend as sppb
from copy import deepcopy

class SparseConv2d(torch.nn.Module):
    def __init__(self, dense_weight, bias=None, padding=0, stride=1, vectorizing_over_on=False):
        super(SparseConv2d, self).__init__()

        self.OC, self.IC, self.K, _ = dense_weight.shape
        self.padding = padding
        self.stride = stride
        self.set_vectorizing_over_on(vectorizing_over_on)

        W_val, W_idx = to_sparse_format_conv2d(dense_weight)
        
        self.W_val = torch.nn.Parameter(W_val)
        self.W_idx = W_idx
        
        assert bias is None or isinstance(bias, torch.nn.Parameter), f"bias is not a parameter but it's {type(bias)}"
        self.bias = bias

    @staticmethod
    def from_dense(conv, vectorizing_over_on=False):
        def bias_to_param():
            if conv.bias is None:
                return None
            return torch.nn.Parameter(conv.bias.data.clone())

        dense_weight = conv.weight.data
        stride = conv.stride[0]
        padding = conv.padding[0]

        return SparseConv2d(
            dense_weight,
            bias=bias_to_param(),
            padding=padding,
            stride=stride,
            vectorizing_over_on=vectorizing_over_on
        )

    def set_vectorizing_over_on(self, vectorizing_over_on):
        self.vectorizing_over_on = vectorizing_over_on
        self.vectorizing_bwd_over_on = vectorizing_over_on
        self.vectorizing_fwd_over_on = vectorizing_over_on and self.stride == 1 # stride 2 is not supported over on

    @property
    def weight(self):
        return self.W_val

    def forward(self, input):
        return SparseConvFunction.apply(input, self.W_val, self.W_idx, self.bias, self.OC, self.K, self.padding, self.stride, self.vectorizing_fwd_over_on, self.vectorizing_bwd_over_on)

    @torch.no_grad()
    def apply_further_mask(self, new_mask, input_shape=None, verbose=False):
        new_mask = (new_mask * 1).int()
        W_idx_OC, W_idx_IC, W_idx_X, W_idx_Y = self.W_idx
        new_nnz = torch.sum(new_mask).item()
        W_val_new = torch.zeros(new_nnz).float()
        W_idx_OC_new = torch.zeros_like(W_idx_OC).int()
        W_idx_IC_new = torch.zeros_like(W_idx_IC).type(torch.short)
        W_idx_X_new = torch.zeros(new_nnz).type(torch.uint8)
        W_idx_Y_new = torch.zeros(new_nnz).type(torch.uint8)
        sppb.further_sparsify_conv2d(
            self.OC,
            self.IC,
            W_idx_OC,
            W_idx_IC,
            W_idx_X,
            W_idx_Y,
            self.W_val.data,
            W_idx_OC_new,
            W_idx_IC_new,
            W_idx_X_new,
            W_idx_Y_new,
            W_val_new,
            new_mask
        )

        sp1 = deepcopy(self)
        sp1.W_val = torch.nn.Parameter(W_val_new)
        sp1.W_idx = W_idx_OC_new, W_idx_IC_new, W_idx_X_new, W_idx_Y_new

        sp2 = deepcopy(sp1)
        sp2.set_vectorizing_over_on(not sp1.vectorizing_over_on)

        sp = run_and_choose([sp1, sp2], input_shape=input_shape, verbose=verbose)
        self.W_val = torch.nn.Parameter(W_val_new)
        self.W_idx = W_idx_OC_new, W_idx_IC_new, W_idx_X_new, W_idx_Y_new
        self.set_vectorizing_over_on(sp.vectorizing_over_on)
    
    def __str__(self):
        nnz = len(self.W_val)
        numel = self.OC * self.IC * self.K * self.K
        return f'SparseConv2d([{self.OC}, {self.IC}, {self.K}, {self.K}], sp={1. - nnz/numel:.2f}, nnz={nnz}, s={self.stride}, p={self.padding}, voo={self.vectorizing_over_on})'


from setuptools import setup, find_packages, Extension
import subprocess

subprocess.run(["pip install pybind11"], shell=True)

proc = subprocess.Popen(["python3 -m pybind11 --includes"], stdout=subprocess.PIPE, shell=True)
(out, err) = proc.communicate()
out = out.decode('ascii').strip().split()

setup(
    name='sparseprop',
    version='0.1.8',    
    description='SparseProp: Efficient Sparse Backpropagation for Faster Training of Neural Networks',
    url='https://github.com/IST-DASLab/sparseprop',
    author='Mahdi Nikdan, Tommaso Pegolotti, Eugenia Iofinova, Eldar Kurtic, Dan Alistarh',
    author_email='mahdi.nikdan@ist.ac.at, tommaso.pegolotti@inf.ethz.ch, eugenia.iofinova@ist.ac.at, eldar.kurtic@ist.ac.at, dan.alistarh@ist.ac.at',
    license='Apache License 2.0',
    packages=find_packages(exclude=['tests', 'tests.*']),
    ext_modules=[Extension(
        'backend',
        [
            'sparseprop/backend.cpp',
        ],
        extra_compile_args=['-O3', '-Wall', '-shared', '-std=c++11', '-fPIC', *out, '-march=native', '-fopenmp', '-ffast-math'],
        extra_link_args=['-lgomp']
    )],
    install_requires=[
        'setuptools>=59.0',
        'pybind11>=2.0.0',
        'scipy',
    ],
    include_package_data=True,
    classifiers=[
        "Programming Language :: C++",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
    ],
)


import torch
import math

from sparseprop import backend as sppb

TRANSPOSE_BLOCK_SIZE = 16

class SparseLinearFunction(torch.autograd.Function):

    @staticmethod
    def forward(ctx, inputT, W_val, W_idx, bias, N):
        input_flat_t = inputT.reshape(-1, inputT.shape[-1])
        B, M = input_flat_t.shape
        if B % TRANSPOSE_BLOCK_SIZE == 0 and M % TRANSPOSE_BLOCK_SIZE == 0:  
            input_flat = torch.zeros(M, B)
            sppb.transpose(input_flat_t, input_flat, TRANSPOSE_BLOCK_SIZE)
        else:
            input_flat = input_flat_t.t().contiguous()
        ctx.inputT_shape = inputT.shape

        M, B = input_flat.shape

        W_idx_N, W_idx_M = W_idx

        output = torch.zeros(N, B).float()
        sppb.sparse_linear_vectorized_forward(input_flat, W_idx_N, W_idx_M, W_val, output)

        ctx.save_for_backward(W_val, bias)
        ctx.svd = (input_flat, W_idx_N, W_idx_M)

        if bias is not None:
            output += bias.view(-1, 1)

        if B % TRANSPOSE_BLOCK_SIZE == 0 and N % TRANSPOSE_BLOCK_SIZE == 0:  
            output_t = torch.zeros(B, N)
            sppb.transpose(output, output_t, TRANSPOSE_BLOCK_SIZE)
        else:
            output_t = output.t() # (B, N)
        output_t = output_t.reshape(*ctx.inputT_shape[:-1], N)
        return output_t

    @staticmethod
    def backward(ctx, grad_output_t):
        W_val, bias = ctx.saved_tensors
        input_flat, W_idx_N, W_idx_M = ctx.svd

        grad_output_t = grad_output_t.reshape(-1, grad_output_t.shape[-1]).contiguous()
        B, N = grad_output_t.shape
        if B % TRANSPOSE_BLOCK_SIZE == 0 and N % TRANSPOSE_BLOCK_SIZE == 0:
            grad_output = torch.zeros(N, B)
            sppb.transpose(grad_output_t, grad_output, TRANSPOSE_BLOCK_SIZE)
        else:
            grad_output = grad_output_t.t().contiguous()

        grad_input = torch.zeros_like(input_flat).float().contiguous() # (M, B)
        grad_W_val = torch.zeros_like(W_val).float().contiguous()
        
        sppb.sparse_linear_vectorized_backward(
            input_flat,
            W_idx_N,
            W_idx_M,
            W_val,
            grad_output,
            grad_input,
            grad_W_val
        )

        M = input_flat.shape[0]
        if B % TRANSPOSE_BLOCK_SIZE == 0 and M % TRANSPOSE_BLOCK_SIZE == 0:  
            grad_input_t = torch.zeros(B, M)
            sppb.transpose(grad_input, grad_input_t, TRANSPOSE_BLOCK_SIZE)
        else:
            grad_input_t = grad_input.t() # (B, M)
        grad_input_t = grad_input_t.reshape(ctx.inputT_shape)
        
        grad_bias = None
        if bias is not None:
            grad_bias = grad_output_t.sum([i for i in range(len(grad_output_t.shape) - 1)])

        return grad_input_t, grad_W_val, None, grad_bias, None


class SparseConvFunction(torch.autograd.Function):

    @staticmethod
    def forward(ctx, input, W_val, W_idx, bias, OC, K, padding, stride, vectorizing_fwd_over_on, vectorizing_bwd_over_on):
        orig_input = input
        
        assert stride in [1, 2], 'only strides 1 and 2 are supported'

        B, IC, M, N = orig_input.shape
        OM = math.ceil((M + 2 * padding - K + 1) / stride)
        ON = math.ceil((N + 2 * padding - K + 1) / stride)
        
        if vectorizing_fwd_over_on:
            assert stride == 1 # only stride 1 is supported in this case, for now
            output = torch.zeros(B, OC, OM, ON).float()
            sppb.sparse_conv2d_vectorized_forward_over_on_stride_1(input, *W_idx, W_val, output, K, padding)
        else:
            input = input.permute(1, 2, 3, 0).contiguous()
            output = torch.zeros(OC, OM, ON, B).float()
            if stride == 1:
                sppb.sparse_conv2d_vectorized_forward_stride_1(input, *W_idx, W_val, output, K, padding)
            elif stride == 2:
                sppb.sparse_conv2d_vectorized_forward_stride_2(input, *W_idx, W_val, output, K, padding)

            output = output.permute(3, 0, 1, 2)
        
        if vectorizing_bwd_over_on: # backward needs the original shape
            ctx.save_for_backward(W_val, bias)
            ctx.svd = (orig_input, *W_idx)
        else:
            ctx.save_for_backward(W_val, bias)
            ctx.svd = (input, *W_idx)
        ctx.K, ctx.padding, ctx.stride = K, padding, stride
        ctx.vectorizing_bwd_over_on = vectorizing_bwd_over_on
        
        if bias is not None:
            output += bias.view(1, -1, 1, 1)

        return output

    @staticmethod
    def backward(ctx, grad_output):
        W_val, bias = ctx.saved_tensors
        input, W_idx_OC, W_idx_IC, W_idx_X, W_idx_Y = ctx.svd
        K, padding, stride = ctx.K, ctx.padding, ctx.stride
        
        vectorizing_bwd_over_on = ctx.vectorizing_bwd_over_on
        
        grad_input = torch.zeros_like(input).float()
        grad_W_val = torch.zeros_like(W_val).float()

        assert stride in [1, 2], 'only stride 1 and 2 are supported'
                
        if vectorizing_bwd_over_on:
            if stride == 1:
                sppb.sparse_conv2d_vectorized_backward_over_on_stride_1(input, W_idx_OC, W_idx_IC, W_idx_X, W_idx_Y, W_val, grad_output, grad_input, grad_W_val, K, padding)
            else:
                sppb.sparse_conv2d_vectorized_backward_over_on_stride_2(input, W_idx_OC, W_idx_IC, W_idx_X, W_idx_Y, W_val, grad_output, grad_input, grad_W_val, K, padding)
        else:
            go = grad_output.permute(1, 2, 3, 0).contiguous()
            if stride == 1:
                sppb.sparse_conv2d_vectorized_backward_stride_1(input, W_idx_OC, W_idx_IC, W_idx_X, W_idx_Y, W_val, go, grad_input, grad_W_val, K, padding)
            else:
                sppb.sparse_conv2d_vectorized_backward_stride_2(input, W_idx_OC, W_idx_IC, W_idx_X, W_idx_Y, W_val, go, grad_input, grad_W_val, K, padding)

        grad_bias = None
        if bias is not None:
            grad_bias = grad_output.sum(dim=(0, 2, 3))

        if not vectorizing_bwd_over_on:
            grad_input = grad_input.permute(3, 0, 1, 2)

        return grad_input, grad_W_val, None, grad_bias, None, None, None, None, None, None

